#include <ruby.h>
#include <pnet/il_engine.h>
#include <pnet/il_image.h>

VALUE IL; /* The Module that incoporates all the IL classes */

VALUE cILExecProcess;
VALUE cILObj;
VALUE cILExecThread;

typedef struct
{
  ILObject *object;
  ILExecThread *thread;
} ILObj;


static void ILExecProcess_free(void *p) 
{
  ILExecProcessDestroy((ILExecProcess *)p);
}

static void ILExecThread_free(void *p)
{
  ILExecThreadDestroy((ILExecThread *)p);
}

static void ILObj_free(void *p)
{
  ILExecThreadDestroy(((ILObj *)p)->thread);
  free(((ILObj *)p)->object);
  free((ILObj *)p);
}

static VALUE ILExecProcess_InitLocale(VALUE class)
{
  ILInitLocale();
  return INT2NUM(1);
}

static VALUE
ILExecProcess_ExecInit(VALUE class, VALUE number)
{
  int arg0 = NUM2INT(number);
  ILExecInit(arg0);
  return INT2NUM(1);
}

static VALUE
ILExecProcess_New(VALUE class, VALUE process_no)
{
  VALUE argv[1];
  ILExecProcess *process = ILExecProcessCreate(NUM2INT(process_no));
  VALUE tdata = Data_Wrap_Struct(class, 0, ILExecProcess_free, process);
  argv[0] = process_no;
  rb_obj_call_init(tdata, 1, argv);
  if (!process) 
    {
      return INT2NUM(0);
    }
  return tdata;
}

static VALUE
ILExecProcess_create(VALUE self, VALUE process_no) 
{
  return ILExecProcess_New(self, process_no);
}

static VALUE
ILExecProcess_process_no(VALUE self)
{
  return rb_iv_get(self, "@process_no");
}

static VALUE
ILExecProcess_init(VALUE self, VALUE process_no)
{
  rb_iv_set(self, "@process_no", process_no);
  rb_iv_set(self, "@library_dirs", rb_arr_new());
  return self;
}

static VALUE
ILExecProcess_library_dirs(VALUE self)
{
  return rb_iv_get(self, "@library_dirs");
}

static VALUE
ILExecProcess_SetLibraryDirs(VALUE self, VALUE list, VALUE rcount) 
{
  int count = NUM2INT(rcount);
  int i=0;
  char **data=NULL;
  ILExecProcess *process;
  if (count > RARRAY(list)->capa) 
    {
      return INT2NUM(0);
    }
  data = (char**)calloc(sizeof(char*), count);
  for(i=0;i<count;i++) 
    {
      if(TYPE(RARRAY(list)->ptr + i) == T_STRING)
	{
	  return INT2NUM(0);
	}
      data[i] = STR2CSTR(RARRAY(list)->ptr + i);
    }
  Data_Get_Struct(self, ILExecProcess, process);
  ILExecProcessSetLibraryDirs(process, data, count);
  rb_iv_set(self, "@library_dirs", list);
  free(data);
  return INT2NUM(1);
}

static VALUE
ILExecProcess_LoadFile(VALUE self, VALUE str)
{
  char *cstr = STR2CSTR(str);
  ILExecProcess *process;
  Data_Get_Struct(self, ILExecProcess, process);
  return INT2NUM(ILExecProcessLoadFile(process, cstr));
}

static VALUE
ILExecProcess_GetMain(VALUE self)
{
  ILExecProcess *process;
  ILExecThread *thread;
  VALUE thread_obj;
  Data_Get_Struct(self, ILExecProcess, process);
  if(!process)
    {
      return INT2NUM(0);
    }
  thread = ILExecProcessGetMain(process);
  if(!thread)
    {
      return INT2NUM(0);
    }
  return Data_Wrap_Struct(cILExecThread, 0, ILExecThread_free, thread);
}

static VALUE
ILObject_ToString(VALUE self)
{
  ILString *val=0;
  ILObj *obj;
  Data_Get_Struct(self, ILObj, obj);
  if (obj->object)
    {
      ILExecThreadCallNamedVirtual(obj->thread, "System.Object", "ToString", "(T)oSystem.String;", &val, obj->object);
    }
  return rb_str_new2(ILStringToAnsi(obj->thread, val));
}

static VALUE
ILObj_New(VALUE class, VALUE thread, VALUE name, VALUE sig, VALUE rparam1, VALUE rparam2)
{
  VALUE argv[5];
  ILExecThread *t;
  char param1;
  int param2;
  char *format = STR2CSTR(sig);
  if (strcmp("(Tci)V", format))
  {
    printf("Oops! forgot to implement VAR_ARG stuff!");
    return INT2NUM(0);
  }
  Data_Get_Struct(thread, ILExecThread, t);
  param1 = STR2CSTR(rparam1)[0]; /* char */
  param2 = NUM2INT(rparam2); /* int */
  ILObj *o; 
  o->object=ILExecThreadNew(t, STR2CSTR(name), format, param1, param2);
  o->thread=t->thread;
  if(!o->object)
    {
      return INT2NUM(0);
    }
  VALUE tdata = Data_Wrap_Struct(class, 0, ILObj_free, o);
  argv[0] = thread;
  argv[1] = name;
  argv[2] = sig;
  argv[3] = rparam1;
  argv[4] = rparam2;
  rb_obj_call_init(tdata, 5, argv);
  return tdata;
}
  
static VALUE
ILObj_init(VALUE self, VALUE thread, VALUE name, VALUE sig, VALUE rparam1, VALUE rparam2)
{
  rb_iv_set(self, "@thread", thread);
  rb_iv_set(self, "@name", name);
  rb_iv_set(self, "@sig", sig);
  rb_iv_set(self, "@rparam1", rparam1);
  rb_iv_set(self, "@rparam2", rparam2);
  return self;
}

static VALUE
ILExecThread_name(VALUE self)
{
  return rb_iv_get(self, "@name");
}

static VALUE
ILExecThread_sig(VALUE self)
{
  return rb_iv_get(self, "@sig");
}

static VALUE
ILExecThread_rparam1(VALUE self)
{
  return rb_iv_get(self, "@rparam1");
}

static VALUE
ILExecThread_rparam2(VALUE self)
{
  return rb_iv_get(self, "@rparam2");
}

void
Init_Pnet()
{
  /* Modules */
  IL = rb_define_module("IL");
  /* Classes */
  cILExecProcess = rb_define_class_under(IL, "Process", rb_cObject);
  cILObj = rb_define_class_under(IL, "Obj", rb_cObject);
  /* IL.Process Methods */
  rb_define_singleton_method(cILExecProcess, "new", ILExecProcess_New, 1);
  rb_define_method(cILExecProcess, "get_main", ILExecProcess_GetMain, 0);
  rb_define_singleton_method(cILExecProcess, "create", ILExecProcess_New, 1);
  rb_define_method(cILExecProcess, "initialize", ILExecProcess_init, 1);
  rb_define_singleton_method(cILExecProcess, "initLocale", ILExecProcess_InitLocale, 0);
  rb_define_method(cILExecProcess, "process_no", ILExecProcess_process_no, 0);
  rb_define_singleton_method(cILExecProcess, "execInit", ILExecProcess_ExecInit, 1);
  rb_define_method(cILExecProcess, "setLibraryDirs", ILExecProcess_SetLibraryDirs,2);
  rb_define_method(cILExecProcess, "library_dirs", ILExecProcess_library_dirs, 0);
  rb_define_method(cILExecProcess, "load", ILExecProcess_LoadFile, 2);
  /* IL.Obj methods */
  rb_define_singleton_method(cILObj, "new", ILObj_New, 5);
  rb_define_method(cILObj, "initialize", ILObj_init, 5);
  rb_define_method(cILObj, "to_s", ILObject_ToString, 0);
}














